/*
 * Copyright 1992 Mike Haertel
 * You may use and redistribute this code subject to the GNU Project's
 * Library General Public License.
 */

#include <errno.h>
#include <float.h>
#include <limits.h>
#include <math.h>
#include <stddef.h>
#include "bignum.h"
#include "strtod.h"

/*
 * Maximum number of Suints ever required to represent a bignum.
 * The fuzz breaks down as 1 for overflow space in divide, 1 for
 * overflow when shifting in decimal digits, 1 for overflow when
 * shifting in the fraction, and 1 to distribute overflow between
 * the fraction and integer portions.
 */
#define VSIZE ((MANT_DIG + MAX_EXP - MIN_EXP + SUINT_BIT - 1) / SUINT_BIT + 4)
#define VFRAC ((MANT_DIG - MIN_EXP + SUINT_BIT - 1) / SUINT_BIT + 2)

/*
 * Maximum number of Suints ever required to hold the base 10 ** SUINT_10_DIG
 * representation of the fraction.  Using MANT_DIG is overkill.
 */
#define FSIZE ((2 * MANT_DIG - MIN_10_EXP + SUINT_10_DIG - 1) / SUINT_10_DIG)

#define MIN(A,B) ((A) < (B) ? (A) : (B))

/*
 * Convert a decimal number represented as a base 10 exponent and
 * a vector of fraction digits to the nearest representable floating
 * point number.  We assume the first digit of input is nonzero.
 */
FLOAT
_STRTOF(int exp, size_t ndig, char digs[])
{
	Suint d, m, v[VSIZE], *vp, f[FSIZE], *fp;
	Duint c;
	int exp2, i, j, nv, nf, nzf, nb, div2, rem;
	FLOAT x;

	if (exp > MAX_10_EXP + 1) {
		errno = ERANGE;
		return HUGE_VAL;
	}

	exp2 = 0;
	i = 0;
	nv = 1;
	vp = v + VFRAC;
	*vp = 0;

	/*
	 * First, get the integer part, in bites of up to SUINT_10_DIG.
	 */
	if (exp > 0) {
		while (exp > 0 && i < ndig) {
			j = MIN(exp, MIN(ndig - i, SUINT_10_DIG));
			exp -= j;
			m = _bigpow10[j];
			j += i;
			d = digs[i++];
			while (i < j)
				d = 10 * d + digs[i++];
			nv = _bigmuladd(vp, nv, vp, m, d);
		}
		while (exp > 0) {
			j = MIN(SUINT_10_DIG, exp);
			m = _bigpow10[j];
			nv = _bigmuladd(vp, nv, vp, m, 0);
			exp -= j;
		}
		nb = _bignbits(nv, vp);
	} else
		nb = 0;

	/*
	 * Now, add in the fraction part, if any, and round to MANT_DIG
	 * bits.  First, we build a base 10 ** SUINT_10_DIG representation
	 * of the fraction.  Then we repeatedly multiply both this fraction
	 * and our fixed point number so far by 2 ** SUINT_BIT, and add
	 * the integer overflow in at the bottom of the fixed point number.
	 */
	if (i < ndig) {
		fp = f + FSIZE;
		while (exp <= -SUINT_10_DIG) {
			*--fp = 0;
			exp += SUINT_10_DIG;
		}
		nzf = f + FSIZE - fp;
		j = 0;
		while (exp < 0)
			++exp, ++j;
		d = 0;
		while (i < ndig) {
			if (j == SUINT_10_DIG) {
				*--fp = d;
				d = 0;
				j = 0;
			}
			d = 10 * d + digs[i++];
			++j;
		}
		if (j < SUINT_10_DIG)
			d *= _bigpow10[SUINT_10_DIG - j];
		*--fp = d;
		nf = f + FSIZE - fp;
		nzf = nf - nzf;
		while (nb <= MANT_DIG) {
			c = 0;
			for (j = 0; j < nzf; ++j) {
				c += (Duint) fp[j] << SUINT_BIT;
				fp[j] = c % SUINT_MAX_POW10;
				c /= SUINT_MAX_POW10;
			}
			for (; j < nf && c != 0; ++j) {
				c += (Duint) fp[j] << SUINT_BIT;
				fp[j] = c % SUINT_MAX_POW10;
				c /= SUINT_MAX_POW10;
			}
			nzf = j;
			exp2 -= SUINT_BIT;
			if (nb != 0) {
				nb += SUINT_BIT;
				++nv;
				*--vp = c;
			} else if (c != 0) {
				*vp = c;
				nb = _bignbits(nv, vp);
			}
		}
	} else
		fp = f, nf = 0;

	/*
	 * Now, decide how many bits of significant to keep (and where
	 * to round).
	 */
	if (nb > MANT_DIG) {
		div2 = nb - MANT_DIG;
		exp2 += div2;
		while (exp2 < MIN_EXP - MANT_DIG)
			++div2, ++exp2;
		rem = _bigrempow2(nv, vp, div2);
		nv = _bigdivpow2(vp, nv, vp, div2);
		if (rem == 2)
			for (j = 0; j < nf; ++j)
				if (fp[j] != 0) {
					++rem;
					break;
				}
		if (rem == 3 || rem == 2 && vp[0] % 2 != 0)
			nv = _bigadd(vp, nv, vp, 1);
	}

	x = 0.0;
	while (nv--)
		x = x * ((Duint) 1 << SUINT_BIT) + vp[nv];

	/*
	 * We rely on ldexp() to detect any overflows or underflows.
	 */
	return ldexp(x, exp2);
}
